@@ -442,17 +442,15 @@ def create_stock_menu_item(stock, txt, callback, ag, accel_key, accel_mod):
     item.connect('activate', callback)
     return item
 
-class HarmonicProgressionLabel(gtk.HBox):
-    """
-    This class can parse strings like I-(6,4)V(5,3)-I and can be used
-    as button labels.
-    """
-    def __init__(self, str):
-        gtk.HBox.__init__(self)
-        self.show()
-        self.m_xalign = 0.5
-        self.m_yalign = 0.5
-        self.set_text(str)
+class AlignedHBox(gtk.HBox):
+    def setup_pre(self):
+        self.g_prebox = gtk.HBox()
+        self.g_prebox.set_no_show_all(True)
+        self.pack_start(self.g_prebox)
+    def setup_post(self):
+        self.g_postbox = gtk.HBox()
+        self.g_postbox.set_no_show_all(True)
+        self.pack_start(self.g_postbox)
     def set_alignment(self, xalign, yalign):
         self.m_xalign = xalign
         self.m_yalign = yalign
@@ -465,13 +463,23 @@ class HarmonicProgressionLabel(gtk.HBox):
         elif xalign == 1.0:
             self.g_prebox.show()
             self.g_postbox.hide()
-    def set_text(self, str):
+
+class HarmonicProgressionLabel(AlignedHBox):
+    """
+    This class can parse strings like I-(6,4)V(5,3)-I and can be used
+    as button labels.
+    """
+    def __init__(self, text):
+        AlignedHBox.__init__(self)
+        self.show()
+        self.m_xalign = 0.5
+        self.m_yalign = 0.5
+        self.set_text(text)
+    def set_text(self, text):
         for o in self.get_children():
             o.destroy()
-        self.m_str = str
-        self.g_prebox = gtk.HBox()
-        self.g_prebox.set_no_show_all(True)
-        self.pack_start(self.g_prebox)
+        self.m_str = text 
+        self.setup_pre()
         while self.m_str:
             T, A, B = self.get_next_token()
             if T == 'big' or T == 'err':
@@ -481,9 +489,7 @@ class HarmonicProgressionLabel(gtk.HBox):
             else:
                 assert T == 'one'
                 self.oneline(A)
-        self.g_postbox = gtk.HBox()
-        self.g_postbox.set_no_show_all(True)
-        self.pack_start(self.g_postbox)
+        self.setup_post()
         self.show_all()
         self.set_alignment(self.m_xalign, self.m_yalign)
     def get_next_token(self):
