@@ -33,8 +33,10 @@ import gobject
 from solfege import cfg
 from solfege import dataparser
 from solfege import filesystem
+from solfege import lfmod
 from solfege import mpd
 from solfege import osutils
+from solfege import parsetree as pt
 from solfege import soundcard
 from solfege import utils
 from solfege import uuid
@@ -139,32 +141,6 @@ _keys_to_interval = {
             }
 
 keywords = (
-    # exercise modules
-    'dictation',
-    'rhythm',
-    'harmonicprogressiondictation',
-    'singchord',
-    'singanswer',
-    'chordvoicing',
-    'chord',
-    'compareintervals',
-    'idbyname',
-    'singinterval',
-    'melodicinterval',
-    'harmonicinterval',
-    'idtone',
-    'twelvetone',
-    'identifybpm',
-    'nameinterval',
-    'elembuilder',
-    'rhythmtapping',
-    'rhythmtapping2',
-    'idproperty',
-    'rhythmdictation',
-    'rhythmdictation2',
-    'toneincontext',
-    'tuner',
-    #
     'harmonic',
     'melodic',
     'progression',
@@ -204,15 +180,15 @@ keywords = (
     'subbass', 'bass', 'varbaritone',
     'baritone', 'tenor', 'alto', 'mezzosoprano', 'soprano',
 )
-predef = {
-    'tempo': (60, 4),
-    'yes': True,
-    'no': False,
-}
+predef = {}
 for n in keywords:
     predef[n] = n
 
 def load_text(parser, filename):
+    """
+    The parser object is the Dataparser object that is calling this
+    function.
+    """
     f = open(os.path.join(parser.m_location, filename), 'r')
     s = f.read()
     f.close()
@@ -245,7 +221,7 @@ class LabelObject(object):
         if name == 'cval':
             return (self.m_labeltype, self.m_labeldata)
 
-lessonfile_functions = {
+lessonfile_builtins = {
            #  (need_parser, function)
         '_': (False, dataparser.dataparser_i18n_func),
         '_i': (False, dataparser.dataparser_i18n__i_func),
@@ -263,7 +239,7 @@ lessonfile_functions = {
         'rhythm': (False, lambda m: Rhythm(m)),
         'percussion': (False, lambda m: Percussion(m)),
         'cmdline': (False, lambda m: Cmdline(m)),
-        'csound': (True, lambda parser, orc, sco: CSound(parser, orc, sco)),
+        'csound': (False, lambda orc, sco: CSound(orc, sco)),
         'wavfile': (False, lambda m: Wavfile(m)),
         'midifile': (False, lambda m: Midifile(m)),
         'mp3file': (False, lambda m: Mp3file(m)),
@@ -273,6 +249,9 @@ lessonfile_functions = {
         'pangomarkup': (False, lambda s: LabelObject("pangomarkup", s)),
         'rnc': (False, lambda s: LabelObject("rnc", s)),
         'chordname': (False, lambda s: LabelObject("chordname", s)),
+    'tempo': (60, 4),
+    'yes': True,
+    'no': False,
 }
 
 
@@ -438,7 +417,7 @@ class MpdParsable(MusicBaseClass):
 
         Return a twoline string showing what caused the exception.
         """
-        if getattr(exception, 'm_obj_lineno', False) is not False:
+        if hasattr(exception, 'm_obj_lineno'):
             return "\n".join((self.m_musicdata.split("\n")[exception.m_obj_lineno],
             " " * exception.m_linepos1 + "^" * (exception.m_linepos2 - exception.m_linepos1)))
         return self._mpd_err_context(exception, lessonfile_ref)
@@ -456,7 +435,7 @@ class MpdParsable(MusicBaseClass):
         get_err_context can create an ok error message.
         """
         # These two lines are just for extra checking...
-        assert getattr(exception, '_coords_done', False) == False
+        assert not hasattr(exception, '_coords_done')
         exception._coords_done = True
         if (exception.m_lineno > self.m_musicdata.count("\n") + 1
             or exception.m_lineno == 0):
@@ -591,10 +570,12 @@ class Chord(ChordCommon):
     def get_music_as_notename_string(self, lessonfile_ref):
         return " ".join(self.get_music_as_notename_list(lessonfile_ref))
     def play(self, lessonfile_ref, question):
-        self.__play(lessonfile_ref, question, question.tempo)
-    def play_slowly(self, lessonfile_ref, question):
         self.__play(lessonfile_ref, question,
-                (question.tempo[0] /2, question.tempo[1]))
+                    lessonfile_ref.get_tempo_of_question(question))
+    def play_slowly(self, lessonfile_ref, question):
+        tempo = lessonfile_ref.get_tempo_of_question(question)
+        self.__play(lessonfile_ref, question,
+                (tempo[0] /2, tempo[1]))
     def __play(self, lessonfile_ref, question, tempo):
         assert isinstance(lessonfile_ref, LessonfileCommon)
         instrument = lessonfile_ref.get_instruments(question, 3)
@@ -760,7 +741,7 @@ class Rvoice(VoiceCommon):
             delta = abs(self.get_first_pitch().m_octave_i)
         except mpd.InvalidNotenameException, e:
             delta = 0
-        if getattr(exception, 'm_obj_lineno', False) is not False:
+        if hasattr(exception, 'm_obj_lineno'):
             return "\n".join((self.m_musicdata.split("\n")[exception.m_obj_lineno],
             " " * (exception.m_linepos1 + delta) + "^" * (exception.m_linepos2 - exception.m_linepos1)))
         return self._mpd_err_context(exception, lessonfile_ref)
@@ -964,7 +945,7 @@ class Cmdline(MusicBaseClass):
             raise osutils.RunningExecutableFailed(self.m_musicdata)
 
 class CSound(MusicBaseClass):
-    def __init__(self, parser, orc, sco):
+    def __init__(self, orc, sco):
         MusicBaseClass.__init__(self, (orc, sco))
     def play(self, lessonfile_ref, question):
         outfile = open(os.path.join(self.temp_dir, "in.orc"), 'w')
@@ -1016,7 +997,7 @@ class Music(MpdTransposable):
             " " * first + "^" * (last - first)))
     def complete_to_musicdata_coords(self, lessonfile_ref, exception):
         # These two lines are just for extra checking...
-        assert getattr(exception, '_coords_done', False) == False
+        assert not hasattr(exception, '_coords_done')
         exception._coords_done = True
         pass
 
@@ -1039,7 +1020,7 @@ def parse_test_def(s):
 
 
 class LessonfileCommon(object):
-    def __init__(self):
+    def __init__(self, module_predefs={}):
         self.m_prev_question = None
         # This variable stores the directory the lesson file is located in.
         # We need this to we can find other files relative to this file.
@@ -1047,6 +1028,8 @@ class LessonfileCommon(object):
         self.m_location = "."
         self._idx = None
         self.m_filename = "<STRING>"
+        self.m_translation_re = re.compile("(?P<varname>\w+)\[(?P<lang>[\w_+]+)\]")
+        self.m_module_predefs = module_predefs
     def parse_file(self, filename):
         """Parse the file named filename. Set these variables:
         self.header     a Header instance
@@ -1054,25 +1037,18 @@ class LessonfileCommon(object):
         """
         self.m_location = os.path.split(uri_expand(filename))[0]
         self.m_filename = filename
-        self.parse_string(open(uri_expand(filename), 'rU').read(), really_filename=filename)
+        s = open(uri_expand(filename), 'rU').read()
+        self.parse_string(s, really_filename=filename)
     def parse_string(self, s, really_filename=None):
         """
         See parse_file docstring.
         """
-        self.dataparser = dataparser.Dataparser(predef, lessonfile_functions, ('tempo',))
-        self.dataparser.m_location = self.m_location
-        try:
-            self.dataparser.parse_string(s, really_filename)
-        except LessonfileParseException, e:
-            e.m_nonwrapped_text = self.dataparser._lexer.get_err_context(self.dataparser._lexer.pos - 2)
-            e.m_token = self.dataparser._lexer.m_tokens[self.dataparser._lexer.pos - 2]
-            raise
+        # FIXME what do we need really_filename for? Error messages?
+        self.get_lessonfile(s, really_filename)
+        ####################
+        # Post parse setup #
         self.m_transpose = mpd.MusicalPitch.new_from_notename("c'")
-        self.header = _Header(self.dataparser.header)
-        self.m_globals = self.dataparser.globals
-        self.m_questions = self.dataparser.questions
-        self.blocklists = self.dataparser.blocklists
-        del self.dataparser
+        self.m_questions = self.blocklists.get('question', [])
         for question in self.m_questions:
             question.active = 1
             # FIXMECOMPAT
@@ -1097,11 +1073,32 @@ class LessonfileCommon(object):
         # Some variables does only make sense if we have a music displayer
         if self.header.at_question_start:
             self.header.have_music_displayer = True
+    def get_lessonfile(self, s, really_filename):
+        """
+        This is the parsetree interpreter.
+        """
+        dp = dataparser.Dataparser()
+        dp.m_location = self.m_location
+        try:
+            dp.parse_string(s, really_filename)
+        except LessonfileParseException, e:
+            e.m_nonwrapped_text = dp._lexer.get_err_context(dp._lexer.pos - 2)
+            e.m_token = dp._lexer.m_tokens[dp._lexer.pos - 2]
+            raise
+        d = lessonfile_builtins.copy()
+        d.update(self.m_module_predefs)
+        mod = lfmod.parse_tree_interpreter(dp.tree, d)
+        self.m_globals = mod.m_globals
+        self.blocklists = mod.m_blocklists
+        if 'header' in self.blocklists:
+            self.header = _Header(self.blocklists['header'][0])
+        else:
+            self.header = _Header({})
 
 
 class QuestionsLessonfile(LessonfileCommon):
-    def __init__(self):
-        LessonfileCommon.__init__(self)
+    def __init__(self, module_predefs={}):
+        LessonfileCommon.__init__(self, module_predefs)
         self.m_discards = []
     def select_random_question(self):
         """
@@ -1242,8 +1239,17 @@ class QuestionsLessonfile(LessonfileCommon):
         assert self._idx is not None
         return self.m_questions[self._idx]
     def get_tempo(self):
+        """
+        Return the tempo of the currently selected question
+        """
         assert self._idx is not None
-        return self.m_questions[self._idx].tempo
+        if 'tempo' in self.m_questions[self._idx]:
+            return self.m_questions[self._idx]['tempo']
+        return self.m_globals['tempo']
+    def get_tempo_of_question(self, question):
+        if 'tempo' in question:
+            return question['tempo']
+        return self.m_globals['tempo']
     def get_name(self):
         """
         Return the translated name of the currently selected question.
@@ -1611,7 +1617,7 @@ class IdByNameLessonfile(QuestionsLessonfile, TestSupport):
         # "labelformat = progression" do as advertised.
         if self.header.labelformat == 'progression':
             for question in self.m_questions:
-                question.label = lessonfile_functions["progressionlabel"][1](question.name)
+                question.label = lessonfile_builtins["progressionlabel"][1](question.name)
         self.discard_questions_without_name()
         self.implements_music_displayer_stafflines()
 
@@ -1765,6 +1771,11 @@ class ElembuilderLessonfile(QuestionsLessonfile):
         self.implements_music_displayer_stafflines()
         if 'element' not in self.blocklists:
             self.blocklists['element'] = []
+        self.m_elements = {}
+        for question in self.m_questions:
+            for e in question['elements']:
+                if e['name'] not in self.m_elements:
+                    self.m_elements[e['name']] = e
 
 
 class InfoCache(object):
@@ -1775,6 +1786,8 @@ class InfoCache(object):
     class FileNotLessonfile(InfoCacheException):
         pass
     class FrontPageCache(object):
+        OLD_FORMAT = 1
+        PARSE_ERROR = 2
         def __init__(self):
             self._data = {}
         def get(self, filename, field):
@@ -1789,9 +1802,14 @@ class InfoCache(object):
                 p = frontpage.load_tree(filename)
                 self._data[filename] = {'title': p.m_name,
                                         'mtime': os.path.getmtime(filename)}
-            except Exception, exception:
-                # FIXME this is a hack
-                self._data[filename] = {'title': str(exception).decode(sys.getfilesystemencoding(), 'replace'), 'mtime': None}
+            except frontpage.OldFormatException:
+                self._data[filename] = self.OLD_FORMAT
+            except frontpage.FrontPageException:
+                self._data[filename] = self.PARSE_ERROR
+        def iter_old_format_files(self):
+            for filename, value in self._data.items():
+                if value == self.OLD_FORMAT:
+                    yield filename
     def __init__(self):
         self._data = {}
         self._dir_mtime = {}
@@ -1815,9 +1833,14 @@ class InfoCache(object):
         if not p:
             raise self.FileNotLessonfile(filename)
         try:
+            try:
+                module = p.header['module'].m_name
+            except AttributeError:
+                # If the module name is the same as a word defined in predef
+                module = p.header['module']
             self._data[filename] = {
                 'title': p.header.get('title', 'error: no title in file'),
-                'module': p.header['module'],
+                'module': module,
                 'test': p.header.get('test', None),
                 'test_requirement': p.header.get('test_requirement', None),
                 'mtime': mtime if mtime else os.path.getmtime(uri_expand(filename)),
@@ -1825,6 +1848,7 @@ class InfoCache(object):
         }
         except KeyError:
             logging.debug("InfoCache.parse_file: FileNotLessonfile(%s)", filename)
+            print "file not lessonfile:", filename
             raise self.FileNotLessonfile(filename)
         if not self._data[filename]['title']:
             self._data[filename]['title'] = "error: empty string as title in '%s'" % filename
@@ -1871,6 +1895,9 @@ class InfoCache(object):
                     return True
                 except StopIteration:
                     logging.debug("parse_all_files(...) done.")
+                    import time
+                    print "all files parsed:", time.time() - start_time
+                    pt.Identifier.check_ns = True
                     return False
             gobject.idle_add(on_idle_parse)
         else:
@@ -1926,99 +1953,6 @@ class InfoCache(object):
                     continue
 
 
-class LessonFileManager:
-    class UnknownLessonID(Exception):
-        pass
-    def __init__(self):
-        # A list of info about lesson files that where discarded
-        # by the lesson file manager
-        self.m_discards = []
-        self.m_uiddb = {}
-    def parse(self, debug):
-        lessonpath = [os.path.join('exercises', 'standard', 'lesson-files'),
-                      filesystem.user_lessonfiles()]
-        if debug:
-            lessonpath.append(os.path.join('exercises', 'standard', 'regression-lesson-files'))
-        self.m_uiddb = {}
-        for dir in lessonpath:
-            if not os.path.isdir(dir):
-                continue
-            v = glob.glob(os.path.join(dir, "*"))
-            for filename in v:
-                filename = filename.decode(locale.getpreferredencoding())
-                # since I usually run solfege from the source dir:
-                if os.path.split(filename)[-1] in ('.arch-ids', 'Makefile', 'Makefile.in') \
-                  or filename.endswith("~") \
-                  or not os.path.isfile(filename):
-                    continue
-                lesson_id = self.parse_into_uiddb(filename)
-                # parse_into_uiddb will return None on error.
-                if lesson_id:
-                    yield lesson_id, self.m_uiddb[lesson_id]
-    def parse_into_uiddb(self, filename):
-                """
-                Returns the lesson_id of the parsed file.
-                Append to .m_discards and return None if there was any errors.
-                """
-                try:
-                    f = LessonIdParser(filename)
-                    if not f.has_lesson_id():
-                        f.add_lesson_id()
-                except IOError, e:
-                    self.m_discards.append({
-                        'filename': filename,
-                        'exception': e,
-                    })
-                    return
-                try:
-                    p = parse_lesson_file_header(filename)
-                except dataparser.DataparserException, e:
-                    return
-                if not p:
-                    return
-                if not 'title' in p.header:
-                    p.header['title'] = '@@@NOT TITLE'
-                if 'module' not in p.header:
-                    self.m_discards.append({
-                        'filename': filename,
-                        'reason': 'no module',
-                    })
-                    return
-                n = {
-                   'filename': filename,
-                   'mtime': os.stat(filename)[stat.ST_MTIME],
-                   'header': p.header,
-                }
-                if not p.header['lesson_id'] in self.m_uiddb:
-                    self.m_uiddb[p.header['lesson_id']] = n
-                else:
-                    # We do the simplest thing here, we keep the oldest
-                    # file and discard info from other files with the
-                    # same lesson_id. Users are in the release notes
-                    # for 3.15.3 advised to solve lesson_id conflict before
-                    # upgrading.
-                    if n['mtime'] < self.m_uiddb[p.header['lesson_id']]['mtime']:
-                        self.m_uiddb[p.header['lesson_id']] = n
-                return p.header['lesson_id']
-    def get(self, lesson_id, fieldname):
-        if lesson_id not in self.m_uiddb:
-            raise self.UnknownLessonID(lesson_id)
-        if fieldname in ('title', 'test', 'module', 'lesson_heading'):
-            if fieldname in self.m_uiddb[lesson_id]['header']:
-                return self.m_uiddb[lesson_id]['header'][fieldname]
-        if fieldname in self.m_uiddb[lesson_id]:
-            return self.m_uiddb[lesson_id][fieldname]
-
-mgr = None
-
-def require_mgr():
-    global mgr
-    if not mgr:
-        mgr = LessonFileManager()
-        # This causes the LessonFileManager to parse all lesson files
-        list(mgr.parse(True))
-
-
 def parse_lesson_file_header(filename):
     """
     This function is used at program starup to get the info the
@@ -2035,70 +1969,17 @@ def parse_lesson_file_header(filename):
     # and must be updated there too if changed.
     s = open(filename, 'rU').read(40960)
     m = r.search(s)
-    p = dataparser.Dataparser(predef, lessonfile_functions)
-    p.m_location = os.path.split(filename)
-    p.m_ignore_lookup_error = True
+    p = LessonfileCommon()
+    ###############################
     if not m:
         return
+    check_ns = pt.Identifier.check_ns
+    pt.Identifier.check_ns = False
     try:
-        p.parse_string(m.group())
+        p.get_lessonfile(m.group(), None)
     except dataparser.DataparserException:
         return None
+    finally:
+        pt.Identifier.check_ns = check_ns    
     return p
 
-
-class LessonIdParser(object):
-    """
-    This is a light weight parser for lesson files that is only used when
-    checking/adding/updating lesson_ids.
-    """
-    def __init__(self, filename):
-        self.m_filename = filename
-        f = open(filename, 'rU')
-        self.m_file_content = f.read()
-        f.close()
-    def has_header_block(self):
-        """
-        Return True if the file has a header block.
-        """
-        return re.search("^\s*header", self.m_file_content, re.MULTILINE) is not None
-    def has_lesson_id(self):
-        """
-        Return True if the lesson file has a lesson_id.
-        """
-        return re.search("lesson_id", self.m_file_content) is not None
-    def add_header_block(self):
-        """
-        Add a header block with a lesson id.
-        """
-        assert not self.has_header_block()
-        v = self.m_file_content.split("\n")
-        for i in range(len(v)):
-            if not v[i].startswith('#'):
-                break
-        self.m_new_id = uuid.generate()
-        v.insert(i+1, "header { \n    lesson_id=\"%s\"\n}" % self.m_new_id)
-        self.m_file_content = "\n".join(v)
-        ofile = open(self.m_filename, 'w')
-        ofile.write(self.m_file_content)
-        ofile.close()
-    def add_lesson_id(self):
-        """
-        Add a lesson id to m_file_content and write it to disk.
-        """
-        assert not self.has_lesson_id()
-        if not self.has_header_block():
-            self.add_header_block()
-        else:
-            ofile = open(self.m_filename, 'w')
-            m = re.search("^\s*header\s*{", self.m_file_content,
-                            re.MULTILINE)
-            if m:
-                ofile.write(self.m_file_content[:m.start()])
-                ofile.write("\nheader {")
-                self.m_new_id = uuid.generate()
-                ofile.write("\n    lesson_id = \"%s\"" % self.m_new_id)
-                ofile.write(self.m_file_content[m.end():])
-            ofile.close()
-
-
