@@ -22,6 +22,7 @@ import hashlib
 import logging
 import os
 import pickle
+import shutil
 import sqlite3
 import time
 
@@ -70,23 +71,34 @@ class DB(object):
                 statistics_filename = ":memory:"
         except NameError:
             statistics_filename = self.get_statistics_filename()
+            # We make a backup of the statistics file since Solfege 3.21.2 will
+            # delete the statistics for the elembuilder exercise.
+            bk_filename = u"%s.pre-3.21.2.backup" % statistics_filename
+            if os.path.exists(statistics_filename) and not os.path.exists(bk_filename):
+                shutil.copyfile(statistics_filename, bk_filename)
             head, tail = os.path.split(statistics_filename)
             if not os.path.exists(head):
                 os.makedirs(head)
-        self.must_recreate = not os.path.exists(statistics_filename)
-        try:
-            if testsuite_is_running:
-                self.must_recreate = False
-        except NameError:
-            pass
         self.conn = sqlite3.connect(statistics_filename)
-        self.setup_tables()
-        # We don't read old data if we are running another profile than
-        # the default one, since there are no old format statistics saved
-        # in the profiles
-        if self.must_recreate and not profile:
-            self.read_old_data(callback)
-        self.upgrade_to_version_2()
+        # Solfege 3.15.0-3.15.2
+        if self.drop_if_has_uuid_column():
+            db_ver = 3
+        elif self.sanity_check():
+            db_ver = 3
+        try:
+            db_ver = self.get_variable("database_version")
+        except self.VariableUndefinedError:
+            # We did not set database_version to 2, ever...
+            db_ver = 1
+        if db_ver < 2:
+            # upgrade_to_version_2() should have deleted the tests table,
+            # but unfortunately it did not. So we cannot test for that
+            # table to check if we are at version 1. But
+            self.upgrade_to_version_2()
+        if db_ver < 3:
+            self.upgrade_to_version_3()
+        self.set_variable("database_version", 3)
+        self.conn.commit()
     def insert_file(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename)
         self.conn.execute("insert into lessonfiles "
@@ -106,24 +118,19 @@ class DB(object):
         self.conn.close()
         os.remove(self.get_statistics_filename())
         self.conn = sqlite3.connect(self.get_statistics_filename())
-        self.setup_tables()
-    def setup_tables(self):
+        self.create_tables()
+    def remove_tables(self):
         """
-        Create needed tables. Drop and recreate the database if it was
-        created by solfege 3.15.0-3.15.2 since we are incompatible with it,
-        and do not want to spend time supporting upgrades from devel
-        releases.
+        Remove the tables if they exist.
         """
-        columns = [x[1] for x in
-            self.conn.execute('pragma table_info(lessonfiles)').fetchall()]
-        # true if the database is created by solfege 3.15.0 - 3.15.2
-        if u'uuid' in columns:
-            logging.debug("statistics: dropping tables because the database is from solfege-3.15.0-3.15.2")
-            self.conn.execute("drop table if exists lessonfiles")
-            self.conn.execute("drop table if exists sessions")
-            self.conn.execute("drop table if exists tests")
-            self.conn.execute("drop table if exists variables")
-            self.must_recreate = False
+        self.conn.execute("drop table if exists sessions")
+        self.conn.execute("drop table if exists sessioninfo")
+        self.conn.execute("drop table if exists lessonfiles")
+        self.conn.execute("drop table if exists variables")
+        # Current Solfege does not use this table, but we did until
+        # the 3.17.0 release. So we remove it if it exists.
+        self.conn.execute("drop table if exists tests")
+    def create_tables(self):
         self.conn.execute(
             "create table if not exists lessonfiles ( "
             " fileid integer primary key autoincrement, "
@@ -144,7 +151,53 @@ class DB(object):
         self.conn.execute("create table if not exists sessioninfo "
             "(fileid int, timestamp int, sessiontype int, "
             "unique (fileid, timestamp))")
-        self.set_variable("database_version", 2)
+        self.set_variable("database_version", 3)
+    def drop_if_has_uuid_column(self):
+        """
+        Drop the database and return True if the lessonfiles table
+        has a uuid column.
+        We started using sqlite to store statistics in Solfege 3.15.0.
+        Solfege 3.15.0-3.15.2 had this column, but it was a bad idea.
+        """
+        columns = [x[1] for x in
+            self.conn.execute('pragma table_info(lessonfiles)').fetchall()]
+        # true if the database is created by solfege 3.15.0 - 3.15.2
+        if u'uuid' in columns:
+            logging.debug("statistics: dropping tables because the database is from solfege-3.15.0-3.15.2")
+            self.remove_tables()
+            self.create_tables()
+            return True
+    def _get_table_names(self):
+        """
+        Return a set of all table names in the database.
+        """
+        tables = self.conn.execute('select name from sqlite_master where type="table";').fetchall()
+        return set([x[0] for x in tables])
+    def sanity_check(self):
+        """
+        The content of the tables 'sessions', 'sessioninfo' and 'lessonfiles'
+        depend of neach other. So we neeed to delete and recreate all tables
+        if any of them are missing. Just to be safe, we delete if the
+        'variables' table are missing tooo.
+
+        A table 'tests' existed until version 3.17, but we don't check for
+        that since we don't use it any more.
+
+        Return True if the tables where deleted and recreated.
+        """
+        tables = self._get_table_names()
+        for t in ('lessonfiles', 'sessions', 'sessioninfo', 'variables'):
+            if t not in tables:
+                missing_table = True
+                break    
+        else:
+            missing_table = False
+        # This is a sanity check. If one of the tables are missing, then
+        # we cannot trust anything, so lets just drop and recreate the database.
+        if missing_table:
+            self.remove_tables()
+            self.create_tables()
+            return True
     def get_fileid(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename), filename
         row = self.conn.execute(
@@ -155,154 +208,47 @@ class DB(object):
             return row[0]
         else:
             raise self.FileNotInDB(filename)
-    def get_uuid_to_filename_mapping(self, callback=None):
+    def get_session_count(self, fileid):
         """
-        Parse all lesson files we can find and return a dict mapping
-        the lesson_id to file names.
-        Current lessonfile manager code cause us to parse the user contributed
-        files in ~/lessonfiles too.
+        Return an int telling the number of practise sessions we have
+        stored for the given fileid.
         """
-        count = 0
-        mapping = {}
-        lessonfile.require_mgr()
-        for uuid, data in lessonfile.mgr.parse(False):
-            mapping[uuid] = lessonfile.mk_uri(data['filename'])
-            count += 1
-            callback(_("Files read: %i") % count)
-        logging.info("get_uuid_to_filename_mapping: scanned %i files.", count)
-        return count, mapping
-    def read_old_data(self, callback):
+        count = self.conn.execute("select count(distinct timestamp) from sessions where fileid=?", (fileid,)).fetchone()[0]
+        return count if count is not None else 0
+    def get_first_timestamp(self, fileid):
         """
-        This function should be run once to import the old format statistics
-        from ~/.solfege/statistics and into the sqlite database.
-        We will import statistics for all uuids that exist in the lessonfiles table.
+        Return the timestamp (as a python datetime object) of the first
+        practise session recorded for this fileid.
+        Return None if no sessions are found.
         """
-        # Snippet from http://wiki.python.org/moin/UsingPickle/RenamingModules
-        # that let us rename even though we have a renamed class
-        renametable = {
-            'src.dataparser': 'solfege.dataparser',
-            }
-
-        def mapname(name):
-            if name in renametable:
-                return renametable[name]
-            return name
-
-        def mapped_load_global(self):
-            module = mapname(self.readline()[:-1])
-            name = mapname(self.readline()[:-1])
-            klass = self.find_class(module, name)
-            self.append(klass)
-
-        def load(file):
-            unpickler = pickle.Unpickler(file)
-            unpickler.dispatch[pickle.GLOBAL] = mapped_load_global
-            return unpickler.load()
-
-        imp, uuid_to_filename = self.get_uuid_to_filename_mapping(callback)
-        def read_session_data(datadir, is_tests, counter):
-            if not os.path.isdir(datadir):
-                logging.debug("read_session_data: '%s' does not exist. Returning", datadir)
-                return
-            for uuid in os.listdir(datadir):
-                if not (os.path.isdir(os.path.join(datadir, uuid))
-                    and os.path.isfile(os.path.join(datadir, '%s_hash' % uuid))):
-                    continue
-                # We do not add statistics for lesson files we don't have
-                # access to.
-                if not uuid in uuid_to_filename:
-                    continue
-                lessonfilename = uuid_to_filename[uuid]
-                try:
-                    fileid = self.get_fileid(lessonfilename)
-                except self.FileNotInDB:
-                    fileid = self.insert_file(lessonfilename)
-                # saved_hashvalue will get the hash value as solfege 3.14 (or
-                # older) calculated when the statistics was saved.
-                try:
-                    s = open(os.path.join(datadir, u"%s_hash" % uuid), "r").read()
-                    try:
-                        saved_hashvalue = int(s)
-                    except ValueError:
-                        saved_hashvalue = None
-                except IOError:
-                    saved_hashvalue = None
-                real_lessonfilename = lessonfile.uri_expand(lessonfilename)
-                s = open(real_lessonfilename, "r").read()
-                # Check if the lesson file in this version of solfege is equal to the
-                # one used when saving the statistics.
-                if hash(s) != saved_hashvalue:
-                    # header.replaces is not used when importing statistics.
-                    # This because the old statistics saved the hash value
-                    # using python standard hash function, and the function
-                    # does not return the same value on all systems.
-                    def bugfix_copy_fn(filename, subdir):
-                        """
-                        Return the filename of the copy of the file if it
-                        exists. Return filename if not.
-                        """
-                        d, f = os.path.split(filename)
-                        fn = os.path.join(subdir, f)
-                        if (d == 'exercises%sstandard%slesson-files' % (os.sep, os.sep)
-                            and os.path.isfile(fn)):
-                            return fn
-                        return filename
-
-                    def bugfix_hash_file(filename, subdir):
-                        copy_fn = bugfix_copy_fn(filename, subdir)
-                        if os.path.isfile(copy_fn):
-                            return hash(open(copy_fn, 'r').read())
-                        return None
-
-                    if not (saved_hashvalue == bugfix_hash_file(real_lessonfilename, "hash-bug-workaround")
-                            or saved_hashvalue == bugfix_hash_file(real_lessonfilename, "hash-bug-workaround2")):
-                        logging.debug("Ignoring statistics for '%s' since the saved hash value does not match.", lessonfilename)
-                        continue
-
-                for timestamp in os.listdir(os.path.join(datadir, uuid)):
-                    if timestamp == u"passed":
-                        continue
-                    f = open(os.path.join(datadir, uuid, timestamp), 'r')
-                    session = load(f)
-                    f.close()
-                    for correct in session.keys():
-                        for guess in session[correct]:
-                            self.conn.execute("insert into sessions "
-                                "(fileid, timestamp, answerkey, guessed, count) "
-                                "values(?, ?, ?, ?, ?)",
-                            (fileid, timestamp, unicode(correct), unicode(guess), session[correct][guess]))
-                    self.conn.execute("insert into sessioninfo "
-                        "(fileid, timestamp, sessiontype) "
-                        "values(?, ?, ?)",
-                        (fileid, timestamp, int(bool(is_tests))))
-                    counter += 1
-                    callback(_("Files read: %i") % counter)
-                if is_tests:
-                    timestamp = self.conn.execute("select timestamp from sessioninfo "
-                        "where fileid=? and sessiontype=1 "
-                        "order by -timestamp ",
-                        (fileid,)).fetchone()[0]
-                    if timestamp:
-                        if lessonfile.infocache.get(lessonfilename, "module") in ('melodicinterval', 'harmonicinterval', 'singinterval'):
-                            parserclass = lessonfile.IntervalsLessonfile
-                        else:
-                            parserclass = lessonfile.IdByNameLessonfile
-                        p = parserclass()
-                        p.parse_file(lessonfilename)
-                        self.cache_new_test_result(lessonfilename,
-                            timestamp,
-                            p.get_test_requirement(),
-                            p.get_test_num_questions())
-            return imp
-        imp = read_session_data(
-                os.path.join(filesystem.app_data(), 'statistics'),
-                False, imp)
-        imp = read_session_data(
-                os.path.join(filesystem.app_data(), 'testresults'),
-                True, imp)
-        self.set_variable("database_version", 2)
-        logging.info("imported statistics for %s lessonfiles", imp)
+        return self.conn.execute("select min(timestamp) from sessions where fileid=?",
+                          (fileid,)).fetchone()[0]
+    def get_last_timestamp(self, fileid):
+        """
+        Return the timestamp (as a python datetime object) of the last
+        practise session recorded for this fileid.
+        Return None if no sessions are found.
+        """
+        return self.conn.execute("select max(timestamp) from sessions where fileid=?",
+                          (fileid,)).fetchone()[0]
+    def delete_statistics(self, filename):
+        fileid = self.get_fileid(filename)
+        cursor = self.conn.cursor()
+        cursor.execute("delete from sessions where fileid=?", (fileid,))
+        cursor.execute("delete from sessioninfo where fileid=?", (fileid,))
         self.conn.commit()
+    def upgrade_to_version_3(self):
+        try:
+            if self.get_variable("database_version") >= 3:
+                return
+        except self.VariableUndefinedError:
+            pass
+        for fn in lessonfile.infocache.iter_parse_all_files():
+            if lessonfile.infocache.get(fn, 'module') == 'elembuilder':
+                try:
+                    self.delete_statistics(fn)
+                except self.FileNotInDB:
+                    pass
     def upgrade_to_version_2(self):
         """
         After releasing 3.16.0 I found that having separate tests and
@@ -315,6 +261,11 @@ class DB(object):
                 return
         except self.VariableUndefinedError:
             pass
+        # This should not happen, but the database_version variable was
+        # not set at the same time as we changed this behaviour, so lets just
+        # return if the table does not exist.
+        if 'test' not in self._get_table_names():
+            return
         # Insert all tests into sessioninfo
         for fileid, timestamp in self.conn.execute("select distinct fileid, timestamp from tests"):
             # This should always succeed since we have just created sessioninfo.
@@ -329,7 +280,7 @@ class DB(object):
         # Insert all practise sessions into sessioninfo
         for fileid, timestamp in self.conn.execute("select distinct fileid, timestamp from sessions"):
             # This can fail because 3.14.11 and earlier had a bug where the same
-            # session was create in both statistics and testresult.
+            # session was created in both statistics and testresult.
             try:
                 self.conn.execute("insert into sessioninfo "
                     "(fileid, timestamp, sessiontype) "
@@ -367,7 +318,6 @@ class DB(object):
                     p.parse_file(filename)
                     self.cache_new_test_result(filename, timestamp,
                         p.get_test_requirement(), p.get_test_num_questions())
-        self.set_variable("database_version", 2)
     def cache_new_test_result(self, filename, timestamp,
                               required, num_questions):
         """
@@ -444,8 +394,7 @@ class DB(object):
                 if not isinstance(replaces, list):
                     replaces = [replaces]
                 if not hashvalue in replaces:
-                    cursor.execute("delete from sessions where fileid=?", (fileid,))
-                    cursor.execute("delete from sessioninfo where fileid=?", (fileid,))
+                    solfege.db.delete_statistics(filename)
                     cursor.execute("update lessonfiles "
                         "set hash=?, test_passed=?, test_result=0.0 where fileid=?",
                         (cur_lessonfile_hash_value, None, fileid))
@@ -467,7 +416,6 @@ class DB(object):
         return {'exercises': different_ex,
                 'practise_count': session_count,
                 'test_count': test_count}
-
     def _recent(self, count, sessiontype):
         """
         sessiontype 0 == normal statistics
@@ -548,7 +496,7 @@ class AbstractStatistics(object):
     def get_keys(self, all_keys=False):
         """
         Return the keys for all questions that have been answered correctly.
-        If all_keys are false, it should also return the correct key for all
+        If all_keys are True, it should also return the correct key for all
         questions that only have been answered wrongly.
         """
         try:
@@ -599,6 +547,15 @@ class AbstractStatistics(object):
         Start a new practise session.
         """
         self.m_timestamp = int(time.time())
+    def reset_custom_mode_session(self, filename):
+        """
+        Exercises that show statistics in custom_mode should call this
+        method instead of reset_session.
+        """
+        self.m_timestamp = 1
+        fileid = solfege.db.get_fileid(filename)
+        solfege.db.conn.execute("delete from sessions "
+                          "where fileid=? and timestamp=1", (fileid,))
     def enter_test_mode(self):
         self.m_test_mode = True
     def exit_test_mode(self):
@@ -719,6 +676,28 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+    def get_num_guess(self, seconds):
+        fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
+        if seconds == -1:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where fileid=?", (fileid,)).fetchone()[0]
+        elif seconds == 0:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where timestamp=? and fileid=?", (self.m_timestamp, fileid,)).fetchone()[0]
+        else:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where timestamp>? and fileid=?", (self.m_timestamp - seconds, fileid)).fetchone()[0]
+        if ret:
+            return ret
+        return 0
+    def get_num_correct(self, seconds):
+        fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
+        if seconds == -1:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where fileid=? and answerkey=guessed", (fileid,)).fetchone()[0]
+        elif seconds == 0:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where timestamp=? and fileid=? and answerkey=guessed", (self.m_timestamp, fileid,)).fetchone()[0]
+        else:
+            ret = solfege.db.conn.execute("select sum(count) from sessions where timestamp>? and fileid=? and answerkey=guessed", (self.m_timestamp - seconds, fileid,)).fetchone()[0]
+        if ret:
+            return ret
+        return 0
     def iter_test_results(self):
         """
         Iterate test results for the associated lesson file, newest
@@ -731,8 +710,8 @@ class AbstractStatistics(object):
         for [timestamp] in solfege.db.conn.execute("select timestamp from sessioninfo where fileid=? and sessiontype=? order by -timestamp", (fileid, 1)):
             ret = {}
             for answerkey, guessed, count in solfege.db.conn.execute("select answerkey, guessed, count from sessions where fileid=? and timestamp=?", (fileid, timestamp)):
-                ret.setdefault(self.int_if_int(answerkey), {})
-                ret[self.int_if_int(answerkey)][self.int_if_int(guessed)] = count
+                ret.setdefault(answerkey, {})
+                ret[answerkey][guessed] = count
             # More necessary than one would expect because we want to handle
             # the possibility that that ret[key1][key2] == None
             # A user has reported that this can happen, but I don't know
